using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using GPUTools.Hair.Scripts.Runtime.Render;
using Battlehub.RTCommon;
using mset;
using Leap.Unity;

namespace Stopper {
	public class LightLayerEditor : MVRScript {

        Dictionary<GameObject, int> originalLayers = new Dictionary<GameObject, int>();

        List<JSONStorableFloat> meshRendererFloats = new List<JSONStorableFloat>();

        HashSet<DAZMergedSkinV2> loadedSkins = new HashSet<DAZMergedSkinV2>();
        HashSet<Component> loadedHairs = new HashSet<Component>();

        HashSet<DAZSkinWrap> replacedClothing = new HashSet<DAZSkinWrap>();

        HashSet<UnityEngine.Object> toDestroy = new HashSet<UnityEngine.Object>();

        JSONStorableFloat skinLayer;
        JSONStorableFloat hairLayer;
        JSONStorableFloat clothingLayer;

        Dictionary<JSONStorableDynamic, JSONStorableDynamic.OnLoaded> handlers = new Dictionary<JSONStorableDynamic, JSONStorableDynamic.OnLoaded>();

        public override void Init() {

            skinLayer = AddSlider("Skin Layer", 28, 0, 31, (layer) => {
                foreach(var skin in loadedSkins){
                    skin.gameObject.layer = (int)layer;
                }
            });

            hairLayer = AddSlider("Hair Layer", 28, 0, 31, (layer) => {
                foreach(var hairRender in loadedHairs){
                    hairRender.gameObject.layer = (int)layer;
                }
            });

            clothingLayer = AddSlider("Clothing Layer", 28, 0, 31, (layer) => {
                foreach(var clothingWrap in replacedClothing){
                    clothingWrap.gameObject.layer = (int)layer;
                }
            });

            var geometry = containingAtom.GetStorableByID("geometry");

            foreach(var character in geometry.GetComponentsInChildren<DAZCharacter>(true)){
                if(character.ready) {
                    RegisterSkin(character);
                } else {
                    JSONStorableDynamic.OnLoaded handler = () => {
                        RegisterSkin(character);
                    };

                    character.onLoadedHandlers += handler;
                    handlers[character] = handler;
                }
            }

            foreach(var hairGroup in geometry.GetComponentsInChildren<DAZHairGroup>(true)){
                if(hairGroup.ready) {
                    StartCoroutine(RegisterHair(hairGroup));
                } else {
                    JSONStorableDynamic.OnLoaded handler = () => {
                        StartCoroutine(RegisterHair(hairGroup));
                    };
                    hairGroup.onLoadedHandlers += handler;
                    handlers[hairGroup] = handler;
                }
            }
            foreach(DAZClothingItem clothingItem in geometry.GetComponentsInChildren<DAZClothingItem>(true)) {
                if(clothingItem.ready) {
                    StartCoroutine(RegisterClothing(clothingItem));
                } else {
                    JSONStorableDynamic.OnLoaded handler = () => {
                        StartCoroutine(RegisterClothing(clothingItem));
                    };
                    
                    clothingItem.onLoadedHandlers += handler;
                    handlers[clothingItem] = handler;
                }
            }
		}

        public void RegisterSkin(DAZCharacter character) {
            foreach(var skin in character.GetComponentsInChildren<DAZMergedSkinV2>(true)) {
                loadedSkins.Add(skin);
                SaveOriginalLayer(skin);
                skin.gameObject.layer = (int)skinLayer.val;
            }
        }

        public IEnumerator RegisterHair(DAZHairGroup hairGroup) {
            // Wait a few frames
            yield return new WaitForSeconds(0.1f);

            foreach(HairRender hairRender in hairGroup.GetComponentsInChildren<HairRender>(true)) {
                loadedHairs.Add(hairRender);
                SaveOriginalLayer(hairRender);
                hairRender.gameObject.layer = (int)hairLayer.val;
            }

            // Also do scalps
            List<DAZSkinWrap> wrapReplacers = ReplaceSkinWrapIfNeeded(hairGroup);

            foreach(var replacedWrap in wrapReplacers) {
                loadedHairs.Add(replacedWrap);
                SaveOriginalLayer(replacedWrap);
                replacedWrap.gameObject.layer = (int)hairLayer.val;
            }
            yield break;
        }

    
        public IEnumerator RegisterClothing(DAZClothingItem clothingItem) {
            // Wait a few frames
            yield return new WaitForSeconds(0.1f);

            List<DAZSkinWrap> wrapReplacers = ReplaceSkinWrapIfNeeded(clothingItem);

            foreach(var replacedWrap in wrapReplacers) {
                replacedClothing.Add(replacedWrap);
                SaveOriginalLayer(replacedWrap);
                replacedWrap.gameObject.layer = (int)clothingLayer.val;
            }
            yield break;
        }

        public List<DAZSkinWrap> ReplaceSkinWrapIfNeeded(Component component){
            List<DAZSkinWrap> wrapReplacers = new List<DAZSkinWrap>();
            bool foundExisting = false;
            foreach(DAZSkinWrap wrap in component.GetComponentsInChildren<DAZSkinWrap>(true)) {
                if($"{wrap.GetType()}".Contains("DAZSkinWrapAccessor")) {
                    wrapReplacers.Add(wrap);
                    wrap.enabled = true;
                    foundExisting = true;
                    break;
                }
            }

            foreach(DAZSkinWrap wrap in component.GetComponentsInChildren<DAZSkinWrap>()) {
                if(!foundExisting) {
                    var newReplacer = wrap.gameObject.AddComponent<DAZSkinWrapReplacer>();
                    wrapReplacers.Add(newReplacer);
                    newReplacer.enabled = true;
                    newReplacer.CopyAttributes(wrap);
                }
                wrap.enabled = false;
            }

            foreach(var replacedWrap in wrapReplacers) {
                var wrapSwitcher = replacedWrap.GetComponent<DAZSkinWrapSwitcher>(); 

                if(wrapSwitcher != null) {
                    // Wrap switchers for custom clothing don't seem to have cloth sim settings set correctly
                    // Calling init fixes that
                    SuperController.LogMessage($"Replacing {replacedWrap} {wrapSwitcher.currentWrapName}");
                    wrapSwitcher.Init();
                    wrapSwitcher.currentWrapName = wrapSwitcher.currentWrapName;
                } else {
                    foreach (DAZSkinWrapMaterialOptions dazskinWrapMaterialOptions in replacedWrap.GetComponents<DAZSkinWrapMaterialOptions>())
                    {
                        dazskinWrapMaterialOptions.skinWrap = replacedWrap;
                    }
                    DAZSkinWrapControl skinWrapControl = base.GetComponent<DAZSkinWrapControl>();
                    if (skinWrapControl != null)
                    {
                        skinWrapControl.wrap = replacedWrap;
                    }
                    foreach (ClothSettings clothSettings in replacedWrap.GetComponents<ClothSettings>())
                    {
                        clothSettings.MeshProvider.PreCalcProvider = replacedWrap;
                    }
                }
            }
            return wrapReplacers;
        }

        public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rightSide = false){
            JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
            storableFloat.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
            uiSlider.rangeAdjustEnabled = false;
            uiSlider.slider.wholeNumbers = true;
            RegisterFloat(storableFloat);
            
            if(onChange != null){
                storableFloat.setCallbackFunction = onChange;
            }

            onChange(defaultVal);
            
            return storableFloat;
	    }

        private void SaveOriginalLayer(Component component) {
            if(!originalLayers.ContainsKey(component.gameObject)){
                originalLayers[component.gameObject] = component.gameObject.layer;
            }
        }
		
        private string GetPathToRoot(GameObject go){
            if (go.transform.parent == null || go.transform.parent == containingAtom.transform)
                return "/" + go.name;
            return GetPathToRoot(go.transform.parent.gameObject) + "/" + go.name;
        }

        private void OnEnable() {
            foreach(var storableFloatName in GetFloatParamNames()){
                GetFloatJSONParam(storableFloatName).setCallbackFunction(GetFloatJSONParam(storableFloatName).val);
            }
        }
		
		private void OnDisable()
		{
			foreach(var entry in originalLayers) {
                if(entry.Key != null) {
                    entry.Key.layer = entry.Value;
                }
            }
		}

        private void OnDestroy() {
            foreach(var obj in toDestroy.ToList()) {
                Destroy(obj);
            }

            foreach(var entry in handlers) {
                if(entry.Key.onLoadedHandlers != null) {
                    entry.Key.onLoadedHandlers -= entry.Value;
                }
            }
        }

        public class EnableListener : MonoBehaviour {
            public Action onEnable;

            private void OnEnable() {
                onEnable();
            }
        }

        public class DAZSkinWrapReplacer : DAZSkinWrap {
            
            public void CopyAttributes(DAZSkinWrap other) {
                this.wrapStore = other.wrapStore;
                this.wrapName = other.wrapName;
                this.draw = other.draw;
                this.debug = other.debug;
                this.debugStartVert = other.debugStartVert;
                this.debugStopVert = other.debugStopVert;
                this.debugTan1 = other.debugTan1;
                this.debugTan2 = other.debugTan2;
                this.debugSize = other.debugSize;
                this.dazMesh = other.dazMesh;
                this.skinTransform = other.skinTransform;
                this.skin = other.skin;
                this.forceRawSkinVerts = other.forceRawSkinVerts;
                this.defaultAdditionalThicknessMultiplier = other.defaultAdditionalThicknessMultiplier;
                this.additionalThicknessMultiplier = other.additionalThicknessMultiplier;
                this.defaultSurfaceOffset = other.defaultSurfaceOffset;
                this.surfaceOffset = other.surfaceOffset;
                this.smoothOuterLoops = other.smoothOuterLoops;
                this.laplacianSmoothPasses = other.laplacianSmoothPasses;
                this.laplacianSmoothBeta = other.laplacianSmoothBeta;
                this.springSmoothPasses = other.springSmoothPasses;
                this.springSmoothFactor = other.springSmoothFactor;
                this.useSpring2 = other.useSpring2;
                this.spring2SmoothFactor = other.spring2SmoothFactor;
                this.moveToSurface = other.moveToSurface;
                this.moveToSurfaceOffset = other.moveToSurfaceOffset;
                this.autoColliderToEnable = other.autoColliderToEnable;
                this.autoColliderToDisable = other.autoColliderToDisable;
                this.forceCPU = other.forceCPU;
                
                this.morphCopyFromTransform = other.morphCopyFromTransform;
                this.morphCopyFrom = other.morphCopyFrom;
                this.vertexIndexLimit = other.vertexIndexLimit;
                this.wrapCheckNormals = other.wrapCheckNormals;
                this.wrapToSkinnedVertices = other.wrapToSkinnedVertices;
                this.wrapToMorphedVertices = other.wrapToMorphedVertices;
                this.wrapToDisabledMaterials = other.wrapToDisabledMaterials;
                this.assetSavePath = other.assetSavePath;


                this.GPUSkinWrapper = other.GPUSkinWrapper;
                this.GPUMeshCompute = other.GPUMeshCompute;

                this.GPUuseSimpleMaterial = other.GPUuseSimpleMaterial;
                this.GPUsimpleMaterial = other.GPUsimpleMaterial;
                this.GPUmaterials = other.GPUmaterials;
                this.simTextures = other.simTextures;
                this.GPUAutoSwapShader = other.GPUAutoSwapShader;
                this.onlyUpdateEnabledMaterials = other.onlyUpdateEnabledMaterials;
                this.materialsEnabled = other.materialsEnabled;
                this.recalculateNormals = other.recalculateNormals;
                this.recalculateTangents = other.recalculateTangents;
                this._numMaterials = this.GPUmaterials.Length;
            }


            // Taken from decompiled VAM code
            private void LateUpdate(){
                if ((this.draw || this.debug) && this.skin != null && !this._updateDrawDisabled && this.dazMesh != null && !this.isWrapping && this.wrapStore != null && this.wrapStore.wrapVertices != null)
                {
                    if (this.autoColliderToEnable != null)
                    {
                        this.autoColliderToEnable.on = true;
                    }
                    if (this.autoColliderToDisable != null)
                    {
                        this.autoColliderToDisable.on = false;
                    }
                    if (this.skin.skinMethod == DAZSkinV2.SkinMethod.CPU || this.forceCPU || !Application.isPlaying)
                    {
                        this.UpdateVerts(false);
                        if (this.draw)
                        {
                            this.DrawMesh();
                        }
                    }
                    else
                    {
                        if (this.skin.delayDisplayOneFrame && this.GPUSkinWrapper != null && this._delayedVertsBuffer != null)
                        {
                            this.GPUSkinWrapper.SetBuffer(this._copyKernel, "inVerts", this._drawVerticesBuffer);
                            this.GPUSkinWrapper.SetBuffer(this._copyKernel, "outVerts", this._delayedVertsBuffer);
                            this.GPUSkinWrapper.Dispatch(this._copyKernel, this.numVertThreadGroups, 1, 1);
                            if (this._normalsBuffer != null)
                            {
                                this.GPUSkinWrapper.SetBuffer(this._copyKernel, "inVerts", this._normalsBuffer);
                                this.GPUSkinWrapper.SetBuffer(this._copyKernel, "outVerts", this._delayedNormalsBuffer);
                                this.GPUSkinWrapper.Dispatch(this._copyKernel, this.numVertThreadGroups, 1, 1);
                            }
                            if (this._tangentsBuffer != null)
                            {
                                this.GPUMeshCompute.SetBuffer(this._copyTangentsKernel, "inTangents", this._tangentsBuffer);
                                this.GPUMeshCompute.SetBuffer(this._copyTangentsKernel, "outTangents", this._delayedTangentsBuffer);
                                this.GPUMeshCompute.Dispatch(this._copyTangentsKernel, this.numVertThreadGroups, 1, 1);
                            }
                        }
                        this.UpdateVertsGPU(true);
                        if (this.draw)
                        {
                            this.DrawMeshGPU(false);
                        }
                    }
                }
            }

            public override void PostProcessDispatch(ComputeBuffer finalVerts)
            {
                if (this.drawInPostProcess && this._wrapVerticesBuffer != null)
                {
                    if (!this.provideToWorldMatrices)
                    {
                        if (this.smoothOuterLoops > 0 && (this.laplacianSmoothPasses > 0 || this.springSmoothPasses > 0))
                        {
                            this.DoSmoothingGPU(finalVerts);
                        }
                        else
                        {
                            this._drawVerticesBuffer = finalVerts;
                        }
                        this.mapVerticesGPU.Map(this._drawVerticesBuffer);
                        this.RunNormalTangentRecalc();
                    }
                    this._updateDrawDisabled = true;
                    this.DrawMeshGPU(true);
                }
            }

            protected void DrawMeshGPU(bool noDelay = false)
            {
                if (!this._renderSuspend)
                {
                    if (this.materialVertsBuffers == null)
                    {
                        this.materialVertsBuffers = new Dictionary<int, ComputeBuffer>();
                    }
                    if (this.materialNormalsBuffers == null)
                    {
                        this.materialNormalsBuffers = new Dictionary<int, ComputeBuffer>();
                    }
                    if (this.materialTangentsBuffers == null)
                    {
                        this.materialTangentsBuffers = new Dictionary<int, ComputeBuffer>();
                    }
                    if (DAZSkinWrap.staticDraw)
                    {
                        Matrix4x4 identity = Matrix4x4.identity;
                        this.dazMesh.DrawMorphedUVMappedMesh(identity);
                    }
                    else if (this.mesh != null)
                    {
                        Matrix4x4 identity2 = Matrix4x4.identity;
                        if (this.dazMesh.drawFromBone != null)
                        {
                            Bounds bounds = this.mesh.bounds;
                            bounds.center = this.dazMesh.drawFromBone.transform.position;
                            Vector3 size = bounds.size;
                            size.x = 0.1f;
                            size.y = 0.1f;
                            size.z = 0.1f;
                            bounds.size = size;
                            this.mesh.bounds = bounds;
                        }
                        for (int i = 0; i < this.mesh.subMeshCount; i++)
                        {
                            if (this.GPUuseSimpleMaterial && this.GPUsimpleMaterial)
                            {
                                if (this.skin.delayDisplayOneFrame && !noDelay)
                                {
                                    this.GPUsimpleMaterial.SetBuffer("verts", this._delayedVertsBuffer);
                                    this.GPUsimpleMaterial.SetBuffer("normals", this._delayedNormalsBuffer);
                                    this.GPUsimpleMaterial.SetBuffer("tangents", this._delayedTangentsBuffer);
                                }
                                else
                                {
                                    this.GPUsimpleMaterial.SetBuffer("verts", this._drawVerticesBuffer);
                                    this.GPUsimpleMaterial.SetBuffer("normals", this._normalsBuffer);
                                    this.GPUsimpleMaterial.SetBuffer("tangents", this._tangentsBuffer);
                                }
                                Graphics.DrawMesh(this.mesh, identity2, this.GPUsimpleMaterial, 0, null, i, null, this.dazMesh.castShadows, this.dazMesh.receiveShadows);
                            }
                            else if (this.materialsEnabled != null && this.materialsEnabled[i] && this.GPUmaterials[i] != null)
                            {
                                if (this.skin != null && this.skin.delayDisplayOneFrame && !noDelay)
                                {
                                    ComputeBuffer computeBuffer;
                                    this.materialVertsBuffers.TryGetValue(i, out computeBuffer);
                                    if (computeBuffer != this._delayedVertsBuffer)
                                    {
                                        this.GPUmaterials[i].SetBuffer("verts", this._delayedVertsBuffer);
                                        this.materialVertsBuffers.Remove(i);
                                        this.materialVertsBuffers.Add(i, this._delayedVertsBuffer);
                                    }
                                    this.materialNormalsBuffers.TryGetValue(i, out computeBuffer);
                                    if (computeBuffer != this._delayedNormalsBuffer)
                                    {
                                        this.GPUmaterials[i].SetBuffer("normals", this._delayedNormalsBuffer);
                                        this.materialNormalsBuffers.Remove(i);
                                        this.materialNormalsBuffers.Add(i, this._delayedNormalsBuffer);
                                    }
                                    this.materialTangentsBuffers.TryGetValue(i, out computeBuffer);
                                    if (computeBuffer != this._delayedTangentsBuffer)
                                    {
                                        this.GPUmaterials[i].SetBuffer("tangents", this._delayedTangentsBuffer);
                                        this.materialTangentsBuffers.Remove(i);
                                        this.materialTangentsBuffers.Add(i, this._delayedTangentsBuffer);
                                    }
                                }
                                else
                                {
                                    ComputeBuffer computeBuffer2;
                                    this.materialVertsBuffers.TryGetValue(i, out computeBuffer2);
                                    if (computeBuffer2 != this._drawVerticesBuffer)
                                    {
                                        this.GPUmaterials[i].SetBuffer("verts", this._drawVerticesBuffer);
                                        this.materialVertsBuffers.Remove(i);
                                        this.materialVertsBuffers.Add(i, this._drawVerticesBuffer);
                                    }
                                    this.materialNormalsBuffers.TryGetValue(i, out computeBuffer2);
                                    if (computeBuffer2 != this._normalsBuffer)
                                    {
                                        this.GPUmaterials[i].SetBuffer("normals", this._normalsBuffer);
                                        this.materialNormalsBuffers.Remove(i);
                                        this.materialNormalsBuffers.Add(i, this._normalsBuffer);
                                    }
                                    this.materialTangentsBuffers.TryGetValue(i, out computeBuffer2);
                                    if (computeBuffer2 != this._tangentsBuffer)
                                    {
                                        this.GPUmaterials[i].SetBuffer("tangents", this._tangentsBuffer);
                                        this.materialTangentsBuffers.Remove(i);
                                        this.materialTangentsBuffers.Add(i, this._tangentsBuffer);
                                    }
                                }
                                Graphics.DrawMesh(this.mesh, identity2, this.GPUmaterials[i], this.gameObject.layer, null, i, null, this.dazMesh.castShadows, this.dazMesh.receiveShadows);
                            }
                        }
                    }
                }
            }
        }
	}
}